import gzip
import re
import subprocess

import dnaio
import pandas as pd
import pysam


def split_fastq_reads(fastq_path, output_path, trim_b=0, size_l=40, size_r=40, size_m=30):
    """
    Split reads in the fastq file into three parts for remapping.
    Depending on the read length, reads may be
    1) skipped,
    2) split into left and right parts,
    3) split into left, right and middle parts

    left size size_l (name with -l suffix)
    right size size_r (name with -r suffix)
    middle size size_m (name with -m suffix)

    Parameters
    ----------
    fastq_path
    output_path
    trim_b
    size_l
    size_r
    size_m

    Returns
    -------

    """
    trim_b = int(trim_b)
    size_max = max(size_l, size_r)
    with dnaio.open(fastq_path) as f, \
            dnaio.open(output_path, mode='w') as out_f:
        for read in f:
            if trim_b > 0:
                read = read[trim_b:-trim_b]
            read_length = len(read)
            if read_length <= size_max:
                continue
            else:
                # split reads to left and right part, they may have overlap
                left_read = read[:size_l]
                left_read.name += '-l'
                out_f.write(left_read)

                right_read = read[-size_r:]
                right_read.name += '-r'
                out_f.write(right_read)

                # if the middle part is longer enough, we also use it
                if read_length >= (size_l + size_r + size_m):
                    middle_read = read[size_l:-size_r]
                    middle_read.name += '-m'
                    out_f.write(middle_read)
    return


def generate_contacts(bam_path, output_path, chrom_size_path, min_gap=1000):
    """
    Assume the input bam file is sort by name,
    1) generate a split table, contain eight columns, corresponding to split parts of R1 and R2
    2) remove duplicates of the split table
    3) generate contacts based on the split table

    Parameters
    ----------
    bam_path
    output_path
    chrom_size_path
    min_gap

    Returns
    -------

    """
    splits = ['1', '1-l', '1-m', '1-r', '2-l', '2-m', '2-r', '2']
    split_type_pattern = re.compile('-[lrm]$')
    split_table_path = f'{output_path}.split.tsv.unsort'

    # check_seq=False, otherwise pysam raise error for empty bam
    with pysam.AlignmentFile(bam_path, check_sq=False) as bam, \
            open(split_table_path, 'w') as out_f:
        # init
        pre_name_base = ''
        record = {s: '' for s in splits}

        for read in bam:
            name = read.query_name
            name_base = name.split('_')[0]
            # pysam always think they are unpaired cause this is SE bam, so I determine manually
            read_type = name.split('_')[1][0]  # read_type is '1' or '2'
            map_to_reverse_strand = read.flag & 16
            split_type = split_type_pattern.findall(name)
            split_type = split_type[0] if len(split_type) == 1 else ''
            split = f'{read_type}{split_type}'

            # reset if name_base changed
            if name_base != pre_name_base:
                out_f.write('\t'.join([record[s] for s in splits]) + '\n')
                pre_name_base = name_base
                record = {s: '' for s in splits}

            if map_to_reverse_strand:
                if read_type == '1':
                    strand = 0
                else:
                    strand = 1
            else:
                if read_type == '1':
                    strand = 1
                else:
                    strand = 0

            if strand:
                pos = f'{strand}:{read.reference_name}:{read.pos + 1}'
            else:
                pos = f'{strand}:{read.reference_name}:{read.pos + read.rlen}'
            record[split] = pos

        # last record
        out_f.write('\t'.join([record[s] for s in splits]) + '\n')

    # remove duplicates by sorting all fields
    cmd = f'sort -k1 -u {split_table_path} -o {split_table_path[:-7]} && rm -f {split_table_path}'
    subprocess.run(cmd, check=True, shell=True)
    split_table_path = split_table_path[:-7]
    _parse_split_table(split_table_path, output_path, chrom_size_path, min_gap=min_gap)
    return


def _parse_split_table(input_path, output_path, chrom_size_path, min_gap=1000):
    """Generate contacts from the split table"""
    chrom_list = pd.read_csv(chrom_size_path, sep='\t', header=None, index_col=0).index.tolist()
    chrom_set = set(chrom_list)
    print('Use these chromosomes in contact file:', chrom_set)
    total = 0
    cis_long = 0
    cis_short = 0
    trans = 0

    # chunk reading reduce memory usage
    data = pd.read_csv(input_path, header=0, sep='\t', chunksize=5000)
    with gzip.open(output_path, 'wt') as f:
        for chunk in data:
            for _, row in chunk.iterrows():
                total += 1
                pos_to_chr = {p: p.split(':')[1]
                              for p in row.dropna()
                              if p.split(':')[1] in chrom_set}
                if len(pos_to_chr) < 2:
                    # only one frag mapped, not contact
                    continue
                else:
                    unique_chroms = set(pos_to_chr.values())
                    if len(unique_chroms) > 2:
                        # more than two chr is abnormal
                        continue
                    elif len(unique_chroms) == 2:
                        # trans contact
                        trans += 1
                        # order pos by chrom list, take one from each side/chrom
                        (read1, _), *_, (read2, _) = sorted(pos_to_chr.items(),
                                                            key=lambda i: chrom_list.index(i[1]))
                        f.write(f'{read1}\t0\t{read2}\t1\n'.replace(':', '\t'))
                    else:
                        # cis contact
                        # order the pos, take left or right most
                        read1, *_, read2 = sorted(pos_to_chr.keys(),
                                                  key=lambda i: int(i.split(':')[-1]))
                        pos1 = int(read1.split(':')[-1])
                        pos2 = int(read2.split(':')[-1])
                        if abs(pos1 - pos2) > min_gap:
                            cis_long += 1
                            f.write(f'{read1}\t0\t{read2}\t1\n'.replace(':', '\t'))
                        else:
                            cis_short += 1
    counts = {'CisShortContact': cis_short,
              'CisLongContact': cis_long,
              'TransContact': trans}
    pd.Series(counts).to_csv(f'{output_path}.counts.txt', header=False)
    return



import pathlib

import pandas as pd


def generate_bigwig(output_dir_path,
                    chrom_size_path,
                    cpu,
                    mc_contexts_list,
                    bin_size):
    output_dir = pathlib.Path(output_dir_path).absolute()
    group_table_path = output_dir / 'GROUP_TABLE.csv'

    group_table = pd.read_csv(group_table_path, index_col=0)
    group_levels = group_table.columns[1:]
    qsub_dir = output_dir / 'qsub'

    bigwig_dir = output_dir / 'BW'
    bigwig_dir.mkdir(exist_ok=True)

    command_records = []
    for group_level in group_levels:
        _group_output_dir = bigwig_dir / group_level
        _group_output_dir.mkdir(exist_ok=True)

        group_dir = output_dir / group_level
        group_allc_records = group_dir / 'ALLC_table.csv'
        group_allc_df = pd.read_csv(group_allc_records, index_col=0)

        for allc_path in group_allc_df['AllcPath']:
            mc_contexts_str = ' '.join(mc_contexts_list)
            output_prefix = _group_output_dir / pathlib.Path(allc_path).name[:-12]

            command = f'allcools allc-to-bigwig ' \
                      f'--allc_path {allc_path} ' \
                      f'--output_prefix {output_prefix} ' \
                      f'--chrom_size_path {chrom_size_path} ' \
                      f'--mc_contexts {mc_contexts_str} ' \
                      f'--bin_size {bin_size} ' \
                      f'--remove_additional_chrom ' \
                      f'--cpu {cpu}'
            command_records.append(command)

    command_file_path = qsub_dir / f'_bigwig.commands.txt'
    with open(command_file_path, 'w') as f:
        f.write('\n'.join(command_records))
    return command_file_path



from collections import defaultdict

import pysam

REVERSE_READ_MCH_CONTEXT = {'CA', 'CC', 'CT'}
FORWARD_READ_MCH_CONTEXT = {'AG', 'TG', 'GG'}


def reverse_complement(seq):
    _seq = ''
    base_map = {'A': 'T',
                'T': 'A',
                'C': 'G',
                'G': 'C'}
    for char in seq.upper()[::-1]:
        _seq += base_map[char]
    return _seq


def single_read_mch_level(read):
    # ref seq is parsed based on read seq and MD tag, and do not depend on reverse or not
    ref_seq = read.get_reference_sequence().upper()
    # use dict instead of string is because ref_seq could contain blocks when skip happen
    ref_seq_dict = {pos: base for pos, base in zip(read.get_reference_positions(), ref_seq)}
    read_seq = read.seq.upper()

    # only count mCH
    mch = 0
    cov = 0
    other_snp = 0
    if read.is_reverse:  # read in reverse strand
        for read_pos, ref_pos, ref_base in read.get_aligned_pairs(matches_only=True, with_seq=True):
            read_base = read_seq[read_pos]
            ref_read_pair = ref_base + read_base
            try:
                ref_context = ref_seq_dict[ref_pos] + ref_seq_dict[ref_pos + 1]
                if ref_context not in REVERSE_READ_MCH_CONTEXT:
                    continue
            except KeyError:
                # ref_seq_dict KeyError means position is on border or not continuous, skip that
                continue
            if ref_read_pair == 'CC':  # C to C means unconverted and methylated
                cov += 1
                mch += 1
            elif ref_read_pair == 'CT':  # C to T means converted and un-methylated
                cov += 1
            else:
                # other kinds of SNPs, do not count to cov
                other_snp += 1
                pass
    else:  # read in forward strand
        for read_pos, ref_pos, ref_base in read.get_aligned_pairs(matches_only=True, with_seq=True):
            read_base = read_seq[read_pos]
            ref_read_pair = ref_base + read_base
            try:
                ref_context = ref_seq_dict[ref_pos - 1] + ref_seq_dict[ref_pos]
                if ref_context not in FORWARD_READ_MCH_CONTEXT:
                    continue
            except KeyError:
                # ref_seq_dict KeyError means position is on border or not continuous, skip that
                continue
            if ref_read_pair == 'GG':  # G to G means unconverted and methylated
                cov += 1
                mch += 1
            elif ref_read_pair == 'GA':  # G to A means converted and un-methylated
                cov += 1
            else:
                # other kinds of SNPs, do not count to cov
                other_snp += 1
                pass
    read_mch_rate = (mch / cov) if cov > 0 else 0
    return read_mch_rate, cov, other_snp


def select_rna_reads(input_bam,
                     output_bam,
                     mc_rate_min_threshold=0.9,
                     cov_min_threshold=3):
    read_profile_dict = defaultdict(int)
    with pysam.AlignmentFile(input_bam) as bam:
        with pysam.AlignmentFile(output_bam, header=bam.header, mode='wb') as out_bam:
            for read in bam:
                read_mch_rate, cov, other_snp = single_read_mch_level(read)
                read_profile_dict[(int(100 * read_mch_rate), cov)] += 1

                # split reads
                if (read_mch_rate < mc_rate_min_threshold) or (cov < cov_min_threshold):
                    continue
                out_bam.write(read)

    with open(str(output_bam) + '.reads_profile.csv', 'w') as stat_f:
        stat_f.write('mc_rate,cov,count\n')
        for (mc_rate, cov), count in read_profile_dict.items():
            stat_f.write(f'{mc_rate},{cov},{count}\n')
    return



from collections import defaultdict

import pysam

METHYLATED_CHAR = 'H'
UNMETHYLATED_CHAR = 'h'


def read_mc_level(bismark_tag):
    m_c = bismark_tag.count(METHYLATED_CHAR)
    normal_c = bismark_tag.count(UNMETHYLATED_CHAR)
    total_c = m_c + normal_c
    if total_c == 0:
        return 0, 0
    else:
        read_mc_rate = m_c / total_c
        return read_mc_rate, total_c


def select_dna_reads(input_bam,
                     output_bam,
                     mc_rate_max_threshold=0.5,
                     cov_min_threshold=3):
    read_profile_dict = defaultdict(int)
    # init dict to make sure the series has something
    read_profile_dict[(50, 50)] = 0
    with pysam.AlignmentFile(input_bam) as f:
        with pysam.AlignmentFile(output_bam, header=f.header, mode='wb') as out_f:
            for read in f:
                bismark_tag = read.get_tag('XM')
                mc_rate, cov = read_mc_level(bismark_tag)
                read_profile_dict[(int(100 * mc_rate), cov)] += 1

                # split reads
                if (mc_rate > mc_rate_max_threshold) or (cov < cov_min_threshold):
                    continue
                out_f.write(read)
    with open(str(output_bam) + '.reads_profile.csv', 'w') as stat_f:
        stat_f.write('mc_rate,cov,count\n')
        for (mc_rate, cov), count in read_profile_dict.items():
            stat_f.write(f'{mc_rate},{cov},{count}\n')
    return



import copy

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.colors import Normalize


def plot_on_plate(data,
                  hue,
                  groupby,
                  ncols=4,
                  plate_base=384,
                  figsize_scale=1,
                  row='Row384',
                  col='Col384',
                  vmin=0,
                  vmax=1,
                  aggregation_func=None):
    """
    Plot metadata into 384 or 96 plate view (heatmap)
    Parameters
    ----------
    data
        dataframe contain plate postion and metric used for color
    hue
        int/float column name used as hue
    groupby
        groupby column, typically groupby plate id column(s) to plot each plate separately
    ncols
        number of column for axes, nrows will be calculated accordingly
    plate_base
        {384, 96} size of the plate view
    figsize_scale
        scale of figure size
    row
        column name for rows
    col
        column name for columns
    vmin
        cmap vmin
    vmax
        cmap vmax
    aggregation_func
        apply to reduce rows after groupby if the row is not unique
    """

    if plate_base == 384:
        plate_nrows, plate_ncols = 16, 24

    elif plate_base == 96:
        plate_nrows, plate_ncols = 8, 12
    else:
        raise ValueError(f'Plate base {plate_base} unknown')

    plot_data_list = []
    plate_names = []
    for plate, sub_df in data.groupby(groupby):
        # check if plate base are duplicated
        duplicated = sub_df[[row, col]].duplicated().sum() != 0
        if duplicated:
            if aggregation_func is None:
                raise ValueError(
                    'Row after groupby is not unique, aggregation_func can not be None'
                )
            plot_data = sub_df.groupby([row,
                                        col])[[hue]].apply(aggregation_func)
        else:
            plot_data = sub_df.set_index([row, col])[[hue]]
        # reindex, missing value will keep as NA
        full_index = pd.MultiIndex.from_tuples([(i, j)
                                                for i in range(plate_nrows)
                                                for j in range(plate_ncols)],
                                               names=[row, col])
        plot_data = plot_data.reindex(full_index).reset_index()
        plot_data_list.append(plot_data)
        if isinstance(plate, str):
            plate_names.append(plate)
        else:
            plate_names.append('\n'.join(plate))

    ncols = min(len(plot_data_list), ncols)
    nrows = int(np.ceil(len(plot_data_list) / ncols))
    cbar_frac = 0.06

    fig = plt.figure(figsize=((6.2 * ncols) * (1 + cbar_frac) * figsize_scale,
                              4 * nrows * figsize_scale))
    gs = fig.add_gridspec(nrows, ncols, wspace=0.1)
    cmap = copy.copy(mpl.cm.get_cmap("viridis"))
    cmap.set_under(color='#440154')
    cmap.set_over(color='#FDE725')
    cmap.set_bad(color='#FFFFFF')
    cnorm = Normalize(vmin, vmax)

    for i, (name, data) in enumerate(zip(plate_names, plot_data_list)):
        ax_row = i // ncols
        ax_col = i % ncols

        ax = fig.add_subplot(gs[ax_row, ax_col])
        ax.scatter(
            x=data[col],
            y=data[row],
            # have to do this, otherwise NaN is skipped.
            c=[cmap(cnorm(v)) for v in data[hue]],
            s=100,
            linewidth=1,
            edgecolor='lightgray')
        ax.set(title=name,
               ylabel='',
               ylim=(plate_nrows, -1),
               yticks=list(range(16)),
               yticklabels=[chr(i + 65) for i in range(0, 16)],
               xlabel='',
               xticks=range(24),
               xticklabels=range(1, 25))
        ax.xaxis.set_tick_params(labelsize=8)
        ax.yaxis.set_tick_params(labelsize=8)
        ax.xaxis.tick_top()
    fig.colorbar(mpl.cm.ScalarMappable(norm=cnorm, cmap=cmap),
                 ax=fig.axes,
                 shrink=0.6,
                 fraction=cbar_frac,
                 label=hue)
    return fig, plate_names, plot_data_list


def cutoff_vs_cell_remain(data,
                          xlim_quantile=(0.01, 0.99),
                          distribution_ylim=None,
                          bins=100,
                          kde=False):
    xlim = tuple(np.quantile(data, xlim_quantile))
    x = np.linspace(xlim[0], xlim[1], 500)
    count_list = np.array([(data > i).sum() for i in x])
    original_total_data = data.size
    count_list = count_list / original_total_data * 100
    data = data[(data < xlim[1]) & (data > xlim[0])]

    fig, ax1 = plt.subplots(figsize=(6, 3))
    ax1 = sns.distplot(data, bins=bins, kde=kde, ax=ax1)
    ax1.set_xlim(xlim)
    ax1.set_xlabel(data.name)
    if distribution_ylim is not None:
        ax1.set_ylim(*distribution_ylim)

    ax2 = ax1.twinx()
    ax2.plot(x, count_list, linewidth=1, c='grey')
    ax2.set_ylabel('% of Cell Remained')
    ax2.set(ylim=(0, 100), yticks=range(0, 101, 10))
    ax2.grid(c='lightgray', linestyle='--', linewidth=0.5)
    return fig, xlim



import pathlib

import pandas as pd

from ...utilities import parse_mc_pattern


def parse_trim_fastq_stats(stat_path):
    # example trim fastq stats
    """
status	in_reads	in_bp	too_short	too_long	too_many_n	out_reads	w/adapters	qualtrim_bp	out_bp
0	OK	1490	213724	0	0	0	1490	4	0	213712
1	status	in_reads	in_bp	too_short	too_long	too_many_n	out_reads	w/adapters	qualtrim_bp	out_bp
2	OK	1490	213712	0	0	0	1482	0	1300	182546
"""
    *cell_id, read_type = pathlib.Path(stat_path).name.split('.')[0].split('-')
    cell_id = '-'.join(cell_id)
    trim_stats = pd.read_csv(stat_path, sep='\t')
    trim_stats = trim_stats.iloc[[0, 2], :].reset_index()  # skip the duplicated title row

    data = pd.Series({
        f'{read_type}InputReads': trim_stats['in_reads'][0],
        f'{read_type}InputReadsBP': trim_stats['in_bp'][0],
        f'{read_type}WithAdapters': trim_stats['w/adapters'][0],
        f'{read_type}QualTrimBP': trim_stats['qualtrim_bp'][1],
        f'{read_type}TrimmedReads': trim_stats['out_reads'][1],
        f'{read_type}TrimmedReadsBP': trim_stats['out_bp'][1],
        f'{read_type}TrimmedReadsRate': int(trim_stats['out_reads'][1]) / int(trim_stats['in_reads'][0])
    }, name=cell_id)
    return data


def parse_bismark_report(stat_path):
    """
    parse bismark output
    """
    *cell_id, read_type = pathlib.Path(stat_path).name.split('.')[0].split('-')
    cell_id = '-'.join(cell_id)
    term_dict = {
        'Number of alignments with a unique best hit from the different alignments': f'{read_type}UniqueMappedReads',
        'Mapping efficiency': f'{read_type}MappingRate',
        'Sequences with no alignments under any condition': f'{read_type}UnmappedReads',
        'Sequences did not map uniquely': f'{read_type}UnuniqueMappedReads',
        'CT/CT': f'{read_type}OT',
        'CT/GA': f'{read_type}OB',
        'GA/CT': f'{read_type}CTOT',
        'GA/GA': f'{read_type}CTOB',
        "Total number of C's analysed": f'{read_type}TotalC',
        'C methylated in CpG context': f'{read_type}TotalmCGRate',
        'C methylated in CHG context': f'{read_type}TotalmCHGRate',
        'C methylated in CHH context': f'{read_type}TotalmCHHRate'}

    with open(stat_path) as rep:
        report_dict = {}
        for line in rep:
            try:
                start, rest = line.split(':')
            except ValueError:
                continue  # more or less than 2 after split
            try:
                report_dict[term_dict[start]] = rest.strip().split('\t')[0].strip('%')
            except KeyError:
                pass
    return pd.Series(report_dict, name=cell_id)


def parse_deduplicate_stat(stat_path):
    *cell_id, read_type = pathlib.Path(stat_path).name.split('.')[0].split('-')
    cell_id = '-'.join(cell_id)
    try:
        dedup_result_series = pd.read_csv(stat_path, comment='#', sep='\t').T[0]
        rename_dict = {
            'UNPAIRED_READS_EXAMINED': f'{read_type}MAPQFilteredReads',
            'UNPAIRED_READ_DUPLICATES': f'{read_type}DuplicatedReads',
            'PERCENT_DUPLICATION': f'{read_type}DuplicationRate'
        }
        dedup_result_series = dedup_result_series.loc[rename_dict.keys()].rename(rename_dict)

        dedup_result_series[f'{read_type}FinalBismarkReads'] = dedup_result_series[f'{read_type}MAPQFilteredReads'] - \
                                                               dedup_result_series[f'{read_type}DuplicatedReads']
        dedup_result_series.name = cell_id
    except pd.errors.EmptyDataError:
        # if a BAM file is empty, picard matrix is also empty
        dedup_result_series = pd.Series({f'{read_type}MAPQFilteredReads': 0,
                                         f'{read_type}DuplicatedReads': 0,
                                         f'{read_type}FinalBismarkReads': 0,
                                         f'{read_type}DuplicationRate': 0}, name=cell_id)
    return dedup_result_series


def generate_allc_stats(output_dir, config):
    output_dir = pathlib.Path(output_dir).absolute()
    allc_stats_dict = {p.name.split('.')[0]: p for p in output_dir.glob('allc/*count.csv')}

    patterns = config['mc_stat_feature'].split(' ')
    patterns_alias = config['mc_stat_alias'].split(' ')
    pattern_translate = {k: v for k, v in zip(patterns, patterns_alias)}

    # real all cell stats
    total_stats = []
    for cell_id, path in allc_stats_dict.items():
        allc_stat = pd.read_csv(path, index_col=0)
        allc_stat['cell_id'] = cell_id
        total_stats.append(allc_stat)
    total_stats = pd.concat(total_stats)
    cell_genome_cov = pd.Series(total_stats.set_index('cell_id')['genome_cov'].to_dict())
    # aggregate into patterns
    cell_records = []
    for pattern in pattern_translate.keys():
        contexts = parse_mc_pattern(pattern)
        pattern_stats = total_stats[total_stats.index.isin(contexts)]
        cell_level_data = pattern_stats.groupby('cell_id')[['mc', 'cov']].sum()
        cell_level_data['frac'] = cell_level_data['mc'] / cell_level_data['cov']

        # prettify col name
        _pattern = pattern_translate[pattern]
        cell_level_data = cell_level_data.rename(
            columns={'frac': f'{_pattern}Frac',
                     'mc': f'{_pattern}mC',
                     'cov': f'{_pattern}Cov'})
        cell_records.append(cell_level_data)
    final_df = pd.concat(cell_records, axis=1, sort=True).reindex(allc_stats_dict.keys())
    final_df['GenomeCov'] = cell_genome_cov
    final_df.index.name = 'cell_id'
    return final_df
